/*
 *	Copyright (C) 2002,2004 Jan Bobrowski <jb@wizard.ae.krakow.pl>
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	version 2 as published by the Free Software Foundation.
 */

#include <unistd.h>
#include <stdio.h>
#include <png.h>
#include "toico.h" /* useful typedefs and macros */

FILE *in, *out;
char *in_name="<stdin>", *out_name="<stdout>";
unsigned img_nr = 0;

void readn(void *p, int n)
{
	int v = fread(p, 1, n, in);
	if(v < n) {
		if(v<0) err(1, "%s", in_name);
		errx(1, "%s: File truncated", in_name);
	}
}

static void skip(int n)
{
	if(n==0) return;
	if(fseek(in, n, SEEK_CUR) < 0) {
		u8 buf[32];
		if(errno != EBADF) err(1, "%s", in_name);
		while(n > sizeof buf) {
			readn(buf, sizeof buf); n -= sizeof buf;
		} while(n>0);
		readn(buf, n);
	}
}

u16 rd16le() {u16 v; readn(&v, sizeof v); return le16(v);}
u32 rd32le() {u32 v; readn(&v, sizeof v); return le32(v);}

static void read_ico_im(int size);
static void write_png();

struct {
	int w, h, d;
	u8 *im, *mask;
	int transparent;
	int npal;
	png_color pal[256];
} im;

static int verbose, promote;

int main(int argc, char *argv[])
{
	int o, count;

	while((o=getopt(argc, argv, "i:n:vt?")) != -1) switch(o) {
		case 'n':
		case 'i': img_nr = atoi(optarg); break;
		case 'v': verbose = 1; break;
		case 't': promote = 1; break;
		default:
			printf("ico2png [-vt] [-n image-number] [icon.ico [picture.png]]\n");
			return 1;
	}
	argc -= optind;
	in = stdin;
	out = stdout;
	if(argc > 0) {
		char *nm = argv[optind];
		if(strcmp(nm, "-") != 0) {
			in = fopen(nm, "r");
			if(!in) err(1, "%s", nm);
			in_name = nm;
		}
	}
	
	if(rd32le() != 0x00010000) errx(1, "%s: Bad format", in_name);
	count = rd16le();
	if(img_nr>=count)
		errx(1, "%s: There're only %d images in icon", in_name, count);

	skip(img_nr*16 + 8);

	{
		u32 size = rd32le();
		u32 offset = rd32le();
		
		skip(offset - (6 + img_nr*16 + 16));
		read_ico_im(size);
		fclose(in);
	}

	if(argc > 1) {
		char *nm = argv[optind+1];
		out = fopen(nm, "w");
		if(!out) err(1, "%s", nm);
		out_name = nm;
	}

	write_png();
	return 0;
}

/****************************************/

typedef struct {
	u32 size;
	u32 width;
	u32 height;
	u16 planes;
	u16 bit_count;
	u32 compression;
	u32 image_size;
	u32 x_pels_per_meter;
	u32 y_pels_per_meter;
	u32 clr_used;
	u32 clr_important;
} BMP_BitmapInfoHeader;

static void read_pal();
static void read_indexed();
static void read_truecolor();
static int read_mask();
static int find_unused_index();
static void apply_mask(int t);
static void apply_rgb_mask();
static void promote_to_rgb();

void read_ico_im(int size)
{
	BMP_BitmapInfoHeader h;
	if(size < sizeof h) goto bad;
	readn(&h, sizeof h);
	im.w = le32(h.width);
	im.h = le32(h.height);
	im.d = le16(h.bit_count);
	if(im.h&1 || im.w<=0 || im.h<=0) goto bad;
	im.h >>= 1;
	if(verbose) fprintf(stderr, "%u. %ux%ux%u ", img_nr, im.w, im.h, im.d);
	if(im.d>0 && im.d<=8 && !(im.d&im.d-1)) {
		read_pal();
		read_indexed();
		im.transparent = -1;
		if(read_mask()) {
			int n;
			if(promote) goto promote;
			n = find_unused_index();
			if(n<256) {
				im.transparent = n;
				if(verbose) fprintf(stderr, " using %u as transparent index", n);
				if(n>=im.npal) {
					im.npal++;
					if(n >= 1<<im.d)
						im.d++;
				}
				apply_mask(n);
			} else {
promote:
				if(verbose) fprintf(stderr, " promoting to true color");
				promote_to_rgb();
				apply_rgb_mask();
			}
		}
	} else if(im.d==24 || im.d==32) {
		read_truecolor();
		read_mask();
		apply_rgb_mask();
	} else {
		if(verbose) putc('\n', stderr);
		errx(1, "%s[%d]: Unsupported depth %d", in_name, img_nr, im.d);
	}
	if(verbose) putc('\n', stderr);
	return;
bad:
	errx(1, "%s[%u]: Bad format", in_name, img_nr);
}

static void read_pal()
{
	int n = 1<<im.d;
	png_color *d = im.pal;
	u8 c[4];
	im.npal = n;
	do {
		readn(c, sizeof c);
		d->blue = c[0];
		d->green = c[1];
		d->red = c[2];
		d++;
	} while(--n>0);
}

#define PITCH(B) ((B)+31 >> 3 & ~3)

static void read_indexed()
{
	int pitch = PITCH(im.w*im.d);
	u8 mask = (1<<im.d) - 1;
	int y = im.h;
	u8 *buf = malloc(pitch);
	im.im = malloc(im.h*im.w);
	do {
		u8 *d = im.im + im.w*--y;
		u8 *s = buf;
		int x = im.w;
		readn(buf, pitch);
		for(;;) {
			int shift = 8;
			u8 v = *s++;
			do {
				shift -= im.d;
				*d++ = v>>shift & mask;
				if(--x==0) goto nxln;
			} while(shift>0);
		}
nxln:;
	} while(y>0);
	free(buf);
}

static void read_truecolor()
{
	int pitch = PITCH(im.w*im.d);
	int step = im.d >> 3;
	u8 *buf = malloc(pitch);
	int y = im.h;
	im.im = malloc(im.h * 4*im.w);
	do {
		u8 *d = im.im + 4*im.w*--y;
		u8 *s = buf;
		int x = im.w;
		readn(buf, pitch);
		do {
			d[0] = s[2];
			d[1] = s[1];
			d[2] = s[0];
			d[3] = step==4 ? s[3] : 0xFF;
			--x; d += 4; s += step;
		} while(x>0);
	} while(y>0);
	free(buf);
}

static int read_mask()
{
	int pitch = PITCH(im.w);
	u8 *buf = malloc(pitch);
	int ret = 0;
	int y = im.h;

	im.mask = malloc(im.h * im.w);
	do {
		u8 *d = im.mask + im.w*--y;
		u8 *s = buf;
		int x = im.w;
		readn(buf, pitch);
		for(;;) {
			u8 b = *s++, m = 0x80;
			do {
				*d = 0xFF;
				if(b&m) {
					*d = 0;
					ret = 1;
				}
				if(--x==0) goto nxln;
				d++;
				m >>= 1;
			} while(m);
		}
nxln:;
	} while(y>0);
	free(buf);
	return ret;
}

static int find_unused_index()
{
	u8 used[256] = {0};
	u8 *i = im.im;
	u8 *m = im.mask;
	int n = im.w * im.h;
	do {
		if(*m++) used[*i] = 1;
		i++;
	} while(--n);
	for(n=0;n<256;n++)
		if(!used[n])
			break;
	return n;
}

static void apply_mask(int t)
{
	u8 *i = im.im;
	u8 *m = im.mask;
	int n = im.w * im.h;
	do {
		if(!*m) *i = t;
		m++; i++;
	} while(--n);
}

static void apply_rgb_mask()
{
	u8 *i = im.im + 3;
	u8 *m = im.mask;
	int n = im.w * im.h;
	do {
		*i &= *m;
		i += 4; m++;
	} while(--n);
}

static void promote_to_rgb()
{
	int n = im.w * im.h;
	u8 *newim = malloc(4*n);
	u8 *i = im.im;
	u8 *d = newim;
	do {
		int c = *i++;
		d[0] = im.pal[c].red;
		d[1] = im.pal[c].green;
		d[2] = im.pal[c].blue;
		d[3] = 0xFF;
		d += 4;
	} while(--n);
	free(im.im);
	im.im = newim;
	im.d = 32;
}

static void write_png()
{
	png_structp png;
	png_infop inf;
	int n, pitch;
	u8 *p;

	png = png_create_write_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
	if(!png) errx(1, "png_create_write_struct");
	inf = png_create_info_struct(png);
	if(!inf) errx(1, "png_create_info_struct");
	png_init_io(png, out);
	png_set_IHDR(png, inf, im.w, im.h, im.d<=8 ? im.d : 8,
		im.d <= 8 ? PNG_COLOR_TYPE_PALETTE : PNG_COLOR_TYPE_RGB_ALPHA,
		PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT,
		PNG_FILTER_TYPE_DEFAULT);

	pitch = 4*im.w;
	if(im.d <= 8) {
		int t = im.transparent;
		if(t>=0) {
			u8 tdata[256];
			int i;
			for(i=0;i<im.npal;i++) tdata[i] = 0xFF;
			tdata[t] = 0;
			png_set_tRNS(png, inf, tdata, im.npal, 0);
		}
		png_set_PLTE(png, inf, im.pal, im.npal);
		pitch = im.w;
	}
	png_write_info(png, inf);
	png_set_packing(png);
	n = im.h;
	p = im.im;
	do {
		png_write_row(png, p);
		p += pitch;
	} while(--n);
	png_write_end(png, inf);
	png_destroy_write_struct(&png, &inf);
}
