/*
 *	Copyright (C) 2001,2004 Jan Bobrowski <jb@wizard.ae.krakow.pl>
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	version 2 as published by the Free Software Foundation.
 */

#include <string.h>
#include "toico.h"

typedef struct {
	u16 reserved;
	u16 type;
	u16 count;
} ICO_Header;

typedef struct {
	u8 width;
	u8 height;
	u8 color_count;
	u8 reserved;
	u16 planes;
	u16 bit_count;
	u32 bytes_in_res;
	u32 image_offset;
} ICO_IconDirEntry;

typedef struct {
	u32 size;
	u32 width;
	u32 height;
	u16 planes;
	u16 bit_count;
	u32 compression;
	u32 image_size;
	u32 x_pels_per_meter;
	u32 y_pels_per_meter;
	u32 clr_used;
	u32 clr_important;
} BMP_BitmapInfoHeader;

static void writen(void *p, int n, FILE *f)
{
	int v = fwrite(p, 1, n, f);
	if(v < n) err(1, "fwrite");
}

static void fput32be(u32 v, FILE *f)
{
	u32 b = be32(v);
	writen(&b, sizeof b, f);
}

static int roundup(int n, int d)
{
	return n*d+31 >> 3 & ~3;
}

static inline int findidx(struct img *i, u32 c)
{
	int n=0;
	if(transparent(c)) return 0;
	for(;;) {
		if(i->pal[n] == c)
			return n;
		n++;
	}
}

static void outbmap(struct img *img, FILE *f)
{
	u32 *e;
	int d = img->depth;

	e = img->im + img->w*img->h;
	do {
		u32 *p = e - img->w;
		do {
			int s=32;
			u32 b=0;
			do {
				u32 c = findidx(img, *p++);
				s -= d;
				b |= c<<s;
			} while(s > 0 && p < e);
			fput32be(b, f);
		} while(p < e);
		e -= img->w;
	} while(e > img->im);
}

static void outbmap24(struct img *img, FILE *f)
{
	u32 *e = img->im + img->w*img->h;
	do {
		u32 *p = e - img->w;
		do {
			u32 c = *p++;
			putc(c    &0xFF, f);
			putc(c>>8 &0xFF, f);
			putc(c>>16&0xFF, f);
		} while(p < e);
		e -= img->w;
	} while(e > img->im);
}

static void outbmap32(struct img *img, FILE *f)
{
	u32 *e;
	
	e = img->im + img->w*img->h;
	do {
		u32 *p = e - img->w;
		do {
			u32 b = le32(*p++ ^ ALPHA);
			writen(&b, sizeof b, f);
		} while(p < e);
		e -= img->w;
	} while(e > img->im);
}

static void outmask(struct img *img, FILE *f)
{
	u32 *e;
	e = img->im + img->w*img->h;
	do {
		u32 *p = e - img->w;
		do {
			int s=32; u32 b=0;
			do {
				s--;
				if(p >= e) break;
				if(transparent(*p++)) b |= 1<<s;
			} while(s > 0);
			fput32be(b, f);
		} while(p < e);
		e -= img->w;
	} while(e > img->im);
}

static void writepal(struct img *img, FILE *f)
{
#ifndef BE_ORDER
	writen(img->pal, 4<<img->depth, f);
#else
	u32 buf[256], *d = buf, *s = img->pal, *e = d + (1<<img->depth);
	do *d++ = le32(*s++); while(d<e);
	writen(buf, 4<<img->depth, f);
#endif
}

/*
 * Checks presence of real alpha channel
 * Also sets transparent color to black.
 */
static inline int fullalpha(struct img *i)
{
	u32 *p = i->im, *e = IMGEND(i);
	do {
		u32 v = *p++ & ALPHA;
		if(!v) continue;
		if(v != ALPHA) return 1;
		p[-1] = v;
	} while(p < e);
	return 0;
}

/*
 * Truncates alpha channel to 1-bit (mask)
 * Also sets transparent color to black.
 */
static void alphatr(struct img *img)
{
#define BG 0xBBBBBB
#define TR 127
	u32 *p = img->im, *e = IMGEND(img);
	do {
		u32 v = *p++;
		if(v >= TR<<24)
			p[-1] = ALPHA;
		else if(v&ALPHA) {
			int a = v>>24 & 0xFF;
			u8 r = v>>16 & 0xFF;
			u8 g = v>>8  & 0xFF;
			u8 b = v     & 0xFF;
			int ca = a^0xFF;
			r = (r*ca + (BG>>16 & 0xFF)*a) / 0xFF;
			g = (g*ca + (BG>>8 & 0xFF)*a) / 0xFF;
			b = (b*ca + (BG & 0xFF)*a) / 0xFF;
			p[-1] = r<<16 | g<<8 | b;
		}
	} while(p < e);
}

static int index_img(struct img *img)
{
	u32 *e = IMGEND(img);
	u32 *p;
	int n = 0;

	p = img->im;
	do if(transparent(*p++)) {
		img->pal[n++] = 0;
		break;
	} while(p<e);

	p = img->im;
	do {
		u32 c = *p++;
		int i;
		if(transparent(c))
			continue;
		for(i=0; i<n; i++)
			if(img->pal[i] == c)
				goto next;
		if(n == 256)
			return 0;
		img->pal[n++] = c;
next:;
	} while(p < e);
	bzero((void*)&img->pal[n], sizeof(img->pal[n])*(256-n));
	img->npal = n;

	img->depth = 8;
	if(n <= 16) {
		img->depth =  4;
		if(n <= 2)
			img->depth = 1;
	}
	return 1;
}

void writeico(struct img *img, int nimg, FILE *f)
{
	int off, n;

	{
		ICO_Header h;
		h.reserved = 0;
		h.type = le16(1);
		h.count = le16(nimg);
		writen(&h, sizeof h, f);
	}

	off = sizeof(ICO_Header) + nimg*sizeof(ICO_IconDirEntry);

	for(n=0; n<nimg; n++) {
		struct img *i = &img[n];
		ICO_IconDirEntry e;
		int sz, np;

		i->npal = 0;
		i->depth = 24;
		if(opt.xp) {
			if(!fullalpha(i))
				goto index;
			i->depth = 32;
		} else {
			alphatr(i);
index:
			index_img(i);
		}

		np = 0;
		if(i->depth <= 8) np = 1<<i->depth;

		if(opt.verbose) {
			fprintf(stderr,"%-16s %dx%dx%d", i->name, i->w, i->h, i->depth);
			if(i->npal) fprintf(stderr, " colors:%d", i->npal);
			putc('\n', stderr);
		}

		e.width = i->w;
		e.height = i->h;
		e.color_count = (u8)np;
		e.reserved = 0;
		e.planes = le16(1);
		e.bit_count = le16(i->depth);
		sz = sizeof(BMP_BitmapInfoHeader) + 4*np
		   + i->h*(roundup(i->w, i->depth) + roundup(i->w, 1));
		e.bytes_in_res = le32(sz);
		e.image_offset = le32(off);
		writen(&e, sizeof e, f);
		off += sz;
	}
	for(n=0; n<nimg; n++) {
		struct img *i = &img[n];
		BMP_BitmapInfoHeader b = {0};
		b.size = le32(sizeof b);
		b.width = le32(i->w);
		b.height = le32(2 * i->h);
		b.planes = le16(1);
		b.bit_count = le16(i->depth);
		b.image_size = le32(i->h*roundup(i->w, i->depth)); /* ??? */
		/*b.clr_important = le32(i->npal);*/ /*spec says this should be zero*/
		writen(&b, sizeof b, f);
		if(i->depth <= 8) {
			writepal(i, f);
			outbmap(i, f);
		} else if(i->depth == 32)
			outbmap32(i, f);
		else 
			outbmap24(i, f);
		outmask(i, f);
	}
}
